/* (c)2013 ALTIMESH & HighPerformanceComputing Magazine France - All rights reserved */

#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <cstdio>

double g_clockRate = 745000 ;
double g_memoryClockRate = 4000000 ;

#include <sys/time.h>
struct timer
{
	struct timeval _start ;
	struct timeval _stop ;

	void start ()
	{
		::gettimeofday (&_start, 0) ;
	}
	void stop ()
	{
		::gettimeofday (&_stop, 0) ;
	}
	double duration ()
	{
		return (((double)_stop.tv_usec - (double)_start.tv_usec) * 1.0e-6) +
		(double)_stop.tv_sec - (double)_start.tv_sec ;
	}
} ;

timer g_tim ;
void StartTimer()
{
	g_tim.start();
}
double GetTimer()
{
	g_tim.stop();
	return g_tim.duration();
}


int g_device = 0 ;
// #define __ldg noldg

__global__ void initkernel(int count, float* a, float* b)
{
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		a [k] = 1.0f ;
	}
}

__global__ void clearfloat4(float4* b)
{
	float4 accum ;
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}

/*
[INFO] : WPT [1] = Took 0.029523 s
[INFO] : WPT [2] = Took 0.022668 s
[INFO] : WPT [4] = Took 0.034825 s
[INFO] : WPT [8] = Took 0.071312 s
[INFO] : WPT [16] = Took 0.131960 s
[INFO] : WPT [32] = Took 0.264243 s
*/
__global__ void accumkernel(int count, int wpt, const float* a, float* b)
{
	float accum = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		accum +=  a [k] ;
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}

// extern "C" __device__ float4 __ldg(const float4*) ;
__device__ float4 __ldg(const float4* a) ;

__device__ float4 noldg (const float4* a) { return *a ; }

/*
[INFO] : WPT [1] = Took 0.011797 s
[INFO] : WPT [2] = Took 0.022601 s
[INFO] : WPT [4] = Took 0.045675 s
[INFO] : WPT [8] = Took 0.085261 s
[INFO] : WPT [16] = Took 0.076980 s
[INFO] : WPT [32] = Took 0.070527 s
*/
__global__ void accumkernelF16_ldg(int count, const float4* a, float4* b)
{
	float4 accum ; 
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		{
			float4 c = __ldg(&a [k * 4 ]) ;
			float4 d = __ldg(&a [k * 4 + 1]) ;
			float4 e = __ldg(&a [k * 4 + 2]) ;
			float4 f = __ldg(&a [k * 4 + 3]) ;
			accum.x += c.x ; accum.y += c.y ; accum.z += c.z ; accum.w += c.w ;
			accum.x += d.x ; accum.y += d.y ; accum.z += d.z ; accum.w += d.w ;
			accum.x += e.x ; accum.y += e.y ; accum.z += e.z ; accum.w += e.w ;
			accum.x += f.x ; accum.y += f.y ; accum.z += f.z ; accum.w += f.w ;
		}
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}
__global__ void accumkernelF16(int count, const float4* a, float4* b)
{
	float4 accum ; 
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		{
			float4 c = a [k * 4 ] ;
			float4 d = a [k * 4 + 1] ;
			float4 e = a [k * 4 + 2] ;
			float4 f = a [k * 4 + 3] ;
			accum.x += c.x ; accum.y += c.y ; accum.z += c.z ; accum.w += c.w ;
			accum.x += d.x ; accum.y += d.y ; accum.z += d.z ; accum.w += d.w ;
			accum.x += e.x ; accum.y += e.y ; accum.z += e.z ; accum.w += e.w ;
			accum.x += f.x ; accum.y += f.y ; accum.z += f.z ; accum.w += f.w ;
		}
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}
__global__ void accumkernelF8_ldg(int count, const float4* a, float4* b)
{
	float4 accum ; 
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		{
			float4 c = __ldg(&a [k * 2 ]) ;
			float4 d = __ldg(&a [k * 2 + 1]) ;
			accum.x += c.x ; accum.y += c.y ; accum.z += c.z ; accum.w += c.w ;
			accum.x += d.x ; accum.y += d.y ; accum.z += d.z ; accum.w += d.w ;
		}
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}
__global__ void accumkernelF8(int count, const float4* a, float4* b)
{
	float4 accum ; 
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		{
			float4 c = a [k * 2 ] ;
			float4 d = a [k * 2 + 1] ;
			accum.x += c.x ; accum.y += c.y ; accum.z += c.z ; accum.w += c.w ;
			accum.x += d.x ; accum.y += d.y ; accum.z += d.z ; accum.w += d.w ;
		}
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}
__global__ void accumkernelF4_ldg(int count, const float4* a, float4* b)
{
	float4 accum ; 
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		{
			float4 c = __ldg(&a [k]) ;
			accum.x += c.x ; accum.y += c.y ; accum.z += c.z ; accum.w += c.w ;
		}
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}

__global__ void accumkernelF4 (int count, const float4* a, float4* b)
{
	float4 accum ; 
	accum.x = 0.0f ; accum.y = 0.0f ; accum.z = 0.0f ; accum.w = 0.0f ;
	for (int k = threadIdx.x + blockDim.x * blockIdx.x ; k < count ; k += blockDim.x * gridDim.x)
	{
		{
			float4 c = a [k] ;
			accum.x += c.x ; accum.y += c.y ; accum.z += c.z ; accum.w += c.w ;
		}
	}
	b [threadIdx.x + blockDim.x * blockIdx.x] = accum ;
}

void cuda_check_func (cudaError_t er, int line, const char* file)
{
	if (er != cudaSuccess) ::fprintf (stderr, "[%s:%d] -- Cuda error : %d %s\n", file, line, er, ::cudaGetErrorString (er)) ;
}
#define CUDA_CHECK(a) cuda_check_func(a,__LINE__,__FILE__)

double g_bestBWF4 = 0.0 ;
double g_bestBW = 0.0 ;
double g_bestBWF4ldg = 0.0 ;

void runkernel (int blockCount, int threadCount, int wpt, int count, float* da, float* db)
{
	FILE* f = ::fopen ("results.csv", "a");

	StartTimer();

	accumkernel <<< blockCount, threadCount >>> (count / wpt, wpt, da, db) ;
	CUDA_CHECK (::cudaDeviceSynchronize());

	double duration = GetTimer();

	double bandwidth = count * sizeof (float) / duration / (1024.0*1024.0*1024.0) ; // in real GB/s
	
	if (g_bestBW < bandwidth) g_bestBW = bandwidth ;

	cudaDeviceProp props ;

	CUDA_CHECK (::cudaGetDeviceProperties (&props, g_device)) ;
	double peekBW = 2.0 * (double)g_memoryClockRate * 1000.0 * (double)props.memoryBusWidth / (8.0*1024.0*1024.0*1024.0) ;

	::fprintf (stderr, "[INFO] : float -- WPT [%d] = Took %lf s -- Achieved BW = %4.2lf GB/s - optimal = %4.2lf GB/s -- usage = %2.1f percent\n", wpt, duration, bandwidth, peekBW, bandwidth / peekBW * 100.0) ;

	if (f != 0)
	{
		//::fprintf (f, "blocks,threads,workperthread,duration,achievedbw,optimalbw,ratio\n") ;
		::fprintf (f, "%d,%d,%d,%0.6lf,%4.2lf,%4.2lf,%2.1f%c\n",blockCount,threadCount/wpt,wpt,duration,bandwidth,peekBW,bandwidth / peekBW * 100.0);
		::fclose (f);
	}
}

void runkernelfloat4_ldg (int blockCount, int threadCount, int count, float* da, float* db, float* hb)
{
	FILE* f = ::fopen ("results-float4-ldg.csv", "a");


	clearfloat4 <<< blockCount, threadCount >>> ((float4*)db) ;
	CUDA_CHECK (::cudaDeviceSynchronize());
	
	StartTimer();

	// accumkernelF4_ldg <<< blockCount, threadCount >>> (count / 4, (float4*)da, (float4*)db) ;
	// accumkernelF8_ldg <<< blockCount, threadCount >>> (count / 8, (float4*)da, (float4*)db) ;
	accumkernelF8_ldg <<< blockCount, threadCount >>> (count / 8, (float4*)da, (float4*)db) ;
	// accumkernelF16_ldg <<< blockCount, threadCount >>> (count / 16, (float4*)da, (float4*)db) ;
	CUDA_CHECK (::cudaDeviceSynchronize());

	double duration = GetTimer();

	CUDA_CHECK (::cudaMemcpy (hb, db, 4 * blockCount * threadCount * sizeof (float), cudaMemcpyDeviceToHost)) ;

	double bandwidth = count * sizeof (float) / duration / (1024.0*1024.0*1024.0) ; // in real GB/s
	
	cudaDeviceProp props ;

	CUDA_CHECK (::cudaGetDeviceProperties (&props, g_device)) ;
	double peekBW = 2.0 * (double)g_memoryClockRate * 1000.0 * (double)props.memoryBusWidth / (8.0*1024.0*1024.0*1024.0) ;

	double total = 0.0 ;
	for (int k = 0 ; k < blockCount ; ++k)
		for (int j = 0 ; j < threadCount * 4 ; ++j)
			total += hb [j + k * threadCount * 4] ;

	if (g_bestBWF4ldg < bandwidth) g_bestBWF4ldg = bandwidth ;

	::fprintf (stderr, "[INFO] : float4 -- res = %lf -- Took %lf s -- Achieved BW = %4.2lf GB/s - optimal = %4.2lf GB/s -- usage = %2.1f percent\n", total, duration, bandwidth, peekBW, bandwidth / peekBW * 100.0) ;

	if (f != 0)
	{
		//::fprintf (f, "blocks,threads,workperthread,duration,achievedbw,optimalbw,ratio\n") ;
		::fprintf (f, "%d,%d,%d,%0.6lf,%4.2lf,%4.2lf,%2.1f%c\n",blockCount,threadCount,1,duration,bandwidth,peekBW,bandwidth / peekBW * 100.0, '%');
		::fclose (f);
	}
}

void runkernelfloat4 (int blockCount, int threadCount, int count, float* da, float* db, float* hb)
{
	FILE* f = ::fopen ("results-float4.csv", "a");


	clearfloat4 <<< blockCount, threadCount >>> ((float4*)db) ;
	CUDA_CHECK (::cudaDeviceSynchronize());
	

	StartTimer();

	accumkernelF4 <<< blockCount, threadCount >>> (count / 4, (float4*)da, (float4*)db) ;
	// accumkernelF8_ldg <<< blockCount, threadCount >>> (count / 8, (float4*)da, (float4*)db) ;
	// accumkernelF8 <<< blockCount, threadCount >>> (count / 8, (float4*)da, (float4*)db) ;
	// accumkernelF16_ldg <<< blockCount, threadCount >>> (count / 16, (float4*)da, (float4*)db) ;
	CUDA_CHECK (::cudaDeviceSynchronize());

	double duration = GetTimer();

	CUDA_CHECK (::cudaMemcpy (hb, db, 4 * blockCount * threadCount * sizeof (float), cudaMemcpyDeviceToHost)) ;

	double bandwidth = count * sizeof (float) / duration / (1024.0*1024.0*1024.0) ; // in real GB/s
	
	cudaDeviceProp props ;

	CUDA_CHECK (::cudaGetDeviceProperties (&props, g_device)) ;
	double peekBW = 2.0 * (double)g_memoryClockRate * 1000.0 * (double)props.memoryBusWidth / (8.0*1024.0*1024.0*1024.0) ;

	double total = 0.0 ;
	for (int k = 0 ; k < blockCount ; ++k)
		for (int j = 0 ; j < threadCount * 4 ; ++j)
			total += hb [j + k * threadCount * 4] ;

	if (g_bestBWF4 < bandwidth) g_bestBWF4 = bandwidth ;

	::fprintf (stderr, "[INFO] : float4 -- res = %lf -- Took %lf s -- Achieved BW = %4.2lf GB/s - optimal = %4.2lf GB/s -- usage = %2.1f percent\n", total, duration, bandwidth, peekBW, bandwidth / peekBW * 100.0) ;

	if (f != 0)
	{
		//::fprintf (f, "blocks,threads,workperthread,duration,achievedbw,optimalbw,ratio\n") ;
		::fprintf (f, "%d,%d,%d,%0.6lf,%4.2lf,%4.2lf,%2.1f%c\n",blockCount,threadCount,1,duration,bandwidth,peekBW,bandwidth / peekBW * 100.0, '%');
		::fclose (f);
	}
}

int main()
{
	
	CUDA_CHECK (::cudaSetDevice (g_device)) ;

	int blockCount = 16 ;
	int threadCount = 1024 ;

	cudaDeviceProp props ;

	CUDA_CHECK (::cudaGetDeviceProperties (&props, g_device)) ;
	blockCount *= props.multiProcessorCount ;
	::fprintf (stderr, "[INFO] : mem bus width = %d\n", props.memoryBusWidth) ;
	::fprintf (stderr, "[INFO] : mem clock rate = %d\n", props.memoryClockRate) ;
	::fprintf (stderr, "[INFO] : gpu clock rate = %d\n", props.clockRate) ;

	g_memoryClockRate = props.memoryClockRate ;
	g_clockRate = props.clockRate ;

	double floatPerQuery = 2.0 * (double)g_memoryClockRate * (double)props.memoryBusWidth / (double)g_clockRate ;
	::fprintf (stderr, "[INFO] : floats per query = %4.2lf\n", floatPerQuery) ;

    const int arraySize = 128*1024*1024;
	float *da, *db ;

	CUDA_CHECK (::cudaMalloc (&da, arraySize * sizeof(float))) ;
	CUDA_CHECK (::cudaMalloc (&db, 4 * blockCount * threadCount * sizeof(float))) ;

	float *hb ;
	hb = new float [4 * blockCount * threadCount] ;
	
	initkernel <<< 16, 1024 >>> (arraySize, da, db) ;
	CUDA_CHECK (::cudaDeviceSynchronize());

	/*
	FILE* f = ::fopen ("results.csv", "w");
	::fprintf (f, "blocks,threads,workperthread,duration,achievedbw,optimalbw,ratio\n") ;
	::fclose (f);
	f = ::fopen ("results-float4.csv", "w");
	::fprintf (f, "blocks,threads,workperthread,duration,achievedbw,optimalbw,ratio\n") ;
	::fclose (f);*/
	
	for (int bkCount = props.multiProcessorCount ; bkCount <= blockCount ; bkCount += props.multiProcessorCount)
	// for (int wpt = 1 ; wpt <= 8 ; wpt *= 2)
	{
		for (int tc = 32 ; tc <= 1024 ; tc += 8)
		{
			runkernelfloat4_ldg (bkCount, tc, arraySize, da, db, hb) ;
		}
	}

	for (int bkCount = props.multiProcessorCount ; bkCount <= blockCount ; bkCount += props.multiProcessorCount)
	// for (int wpt = 1 ; wpt <= 8 ; wpt *= 2)
	{
		for (int tc = 32 ; tc <= 1024 ; tc += 8)
		{
			runkernelfloat4 (bkCount, tc, arraySize, da, db, hb) ;
		}
	}

	for (int bkCount = props.multiProcessorCount ; bkCount <= blockCount ; bkCount += props.multiProcessorCount)
	// for (int wpt = 1 ; wpt <= 8 ; wpt *= 2)
	{
		for (int tc = 32 ; tc <= 1024 ; tc += 8)
		{
			runkernel (bkCount, tc,  1, arraySize, da, db) ;
		}
	}

	::fprintf (stderr, "[INFO] : float4 LDG -- best achieved bandwidth = %lf\n", g_bestBWF4ldg) ;
	::fprintf (stderr, "[INFO] : float4     -- best achieved bandwidth = %lf\n", g_bestBWF4) ;
	::fprintf (stderr, "[INFO] : float      -- best achieved bandwidth = %lf\n", g_bestBW) ;
}


