/* (c)2013 ALTIMESH & HighPerformanceComputing Magazine France - All rights reserved */

#include <stdio.h>
#include <sys/time.h>
#include "cuda_runtime.h"

struct timer
{
	struct timeval _start ;
	struct timeval _stop ;

	void start ()
	{
		::gettimeofday (&_start, 0) ;
	}
	void stop ()
	{
		::gettimeofday (&_stop, 0) ;
	}
	double duration ()
	{
		return (((double)_stop.tv_usec - (double)_start.tv_usec) * 1.0e-6) +
		(double)_stop.tv_sec - (double)_start.tv_sec ;
	}
} ;

timer g_tim ;
void StartTimer()
{
	g_tim.start();
}
double GetTimer()
{
	g_tim.stop();
	return 1000.0 * g_tim.duration();
}

__global__ void initkernel(int arraySize, float * a, float* b, float* c, float* d, double* ad, double* bd, double* cd)
{
    for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < arraySize ; i += blockDim.x * gridDim.x)
       {
             a [i] = 1.0f / (0.5f + (float)i) ;
             b [i] = 0.0f ;
             c [i] = 0.0f ;
             d [i] = 0.0f ;
             ad [i] = 1.0 / (0.5 + (double)i) ;
             bd [i] = 0.0f ;
             cd [i] = 0.0f ;
       }
}

// x*y+z = __fmaf_rn (x,y,z)

// expm1 (x) = x + x2/2! + x3/3! + x4/4! + ...
// = ((((((x + 8) * x + 7*8) * x + 6*7*8) * x + 5*6*7*8) * x +
__device__ __forceinline__ float myexpm1 (float x)
{
       float res = x + 8.0f ;
       res = res * x + 56.0f ;
       res = res * x + 336.0f ;
       res = res * x + 1680.0f ;
       res = res * x + 6720.0f ;
       res = res * x + 20160.0f ;
       return res * x * 2.4801587301587301587301587301587e-5f ;
}

__global__ void expkernel(int arraySize, const float * __restrict a, float* b)
{
       for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < arraySize ; i += blockDim.x * gridDim.x)
       {
             b [i] = myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (myexpm1 (a [i])))))))))))) ;
       }
}

struct vect4
{
       float4 val ;

       __device__ __forceinline__ vect4 () {} ;
       __device__ __forceinline__ vect4 (const float4& v) : val (v) {}

       __device__ __forceinline__ operator float4 () { return val ; }

       __device__ __forceinline__ void self__fmaf4_rn (const vect4& b, float c)
       {
             val.x = ::__fmaf_rn (val.x, b.val.x, c) ;
             val.y = ::__fmaf_rn (val.y, b.val.y, c) ;
             val.z = ::__fmaf_rn (val.z, b.val.z, c) ;
             val.w = ::__fmaf_rn (val.w, b.val.w, c) ;
       }

       static __device__ __forceinline__ vect4 __fmaf4_rn (const vect4& a, const vect4& b, const vect4& c)
       {
             vect4 res ;
             res.val.x = ::__fmaf_rn (a.val.x, b.val.x, c.val.x) ;
             res.val.y = ::__fmaf_rn (a.val.y, b.val.y, c.val.y) ;
             res.val.z = ::__fmaf_rn (a.val.z, b.val.z, c.val.z) ;
             res.val.w = ::__fmaf_rn (a.val.w, b.val.w, c.val.w) ;
             return res ;
       }

       static __device__ __forceinline__ vect4 __fmaf4_rn (const vect4& a, const vect4& b, float c)
       {
             vect4 res ;
             res.val.x = ::__fmaf_rn (a.val.x, b.val.x, c) ;
             res.val.y = ::__fmaf_rn (a.val.y, b.val.y, c) ;
             res.val.z = ::__fmaf_rn (a.val.z, b.val.z, c) ;
             res.val.w = ::__fmaf_rn (a.val.w, b.val.w, c) ;
             return res ;
       }

       __device__ __forceinline__ vect4 operator* (const vect4& a)
       {
             vect4 res ;
             res.val.x = val.x * a.val.x ;
             res.val.y = val.y * a.val.y ;
             res.val.z = val.z * a.val.z ;
             res.val.w = val.w * a.val.w ;
             return res ;
       }

       __device__ __forceinline__ vect4 operator+(float v)
       {
             vect4 res ;
             res.val.x = val.x + v ;
             res.val.y = val.y + v ;
             res.val.z = val.z + v ;
             res.val.w = val.w + v ;
             return res ;
       }

       __device__ __forceinline__ vect4 operator*(float v)
       {
             vect4 res ;
             res.val.x = val.x * v ;
             res.val.y = val.y * v ;
             res.val.z = val.z * v ;
             res.val.w = val.w * v ;
             return res ;
       }
} ;

__device__ vect4 myexpm1vect4 (vect4 x)
{
       vect4 res = x + 8.0f ;
       /* res.self__fmaf4_rn (x, 56.0f) ;
       res.self__fmaf4_rn (x, 336.0f) ;
       res.self__fmaf4_rn (x, 1680.0f) ;
       res.self__fmaf4_rn (x, 6720.0f) ;
       res.self__fmaf4_rn (x, 20160.0f) ; */

       res = vect4::__fmaf4_rn (res, x, 56.0f) ;
       res = vect4::__fmaf4_rn (res, x, 336.0f) ;
       res = vect4::__fmaf4_rn (res, x, 1680.0f) ;
       res = vect4::__fmaf4_rn (res, x, 6720.0f) ;
       res = vect4::__fmaf4_rn (res, x, 20160.0f) ;
       return res * x * 2.4801587301587301587301587301587e-5f ;
}
// __device__ float4 __ldg(const float4* a) ;
__global__ void expkernelvect4(int arraySize, const float4 * a, float4 * b)
{
       for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < arraySize ; i += blockDim.x * gridDim.x)
       {
             // b [i] = a [i] ;
             b [i] = myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (myexpm1vect4 (a [i])))))))))))) ;
       }
}

__device__ __forceinline__ float4 operator+(float4 x, float y)
{
       float4 res ;
       res.x = x.x + y ;
       res.y = x.y + y ;
       res.z = x.z + y ;
       res.w = x.w + y ;
       return res ;
}

__device__ __forceinline__ float4 operator*(float4 x, float y)
{
       float4 res ;
       res.x = x.x * y ;
       res.y = x.y * y ;
       res.z = x.z * y ;
       res.w = x.w * y ;
       return res ;
}

__device__ __forceinline__ float4 operator*(float4 x, float4 y)
{
       float4 res ;
       res.x = x.x * y.x ;
       res.y = x.y * y.y ;
       res.z = x.z * y.z ;
       res.w = x.w * y.w ;
       return res ;
}

__device__ __forceinline__ float4 __fmaf4_rn(float4 x, float4 y, float z)
{
       float4 res ;
       res.x = ::__fmaf_rn (x.x, y.x, z) ;
       res.y = ::__fmaf_rn (x.y, y.y, z) ;
       res.z = ::__fmaf_rn (x.z, y.z, z) ;
       res.w = ::__fmaf_rn (x.w, y.w, z) ;
       return res ;
}

// operations count (each)
// + : 1
// * : 2
// *+ : 5
__device__ __forceinline__ float4 myexpm1float4 (float4 x)
{
       float4 res = x + 8.0f ;
       res = __fmaf4_rn (res, x, 56.0f) ;
       res = __fmaf4_rn (res, x, 336.0f) ;
       res = __fmaf4_rn (res, x, 1680.0f) ;
       res = __fmaf4_rn (res, x, 6720.0f) ;
       res = __fmaf4_rn (res, x, 20160.0f) ;
       return res * x * 2.4801587301587301587301587301587e-5f ;
}

__device__ float4 __ldg(const float4* a) ;
// 64 * (arraysize * 4) flops minimum for this kernel
__global__ void expkernelfloat4(int arraySize, const float4 * a, float4 * b)
{
       for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < arraySize ; i += blockDim.x * gridDim.x)
       {
             // b [i] = a [i] ;
             b [i] = myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (myexpm1float4 (__ldg(&(a [i])))))))))))))) ;
       }
}



__device__ __forceinline__ double4 operator+(double4 x, float y)
{
       double4 res ;
       res.x = x.x + y ;
       res.y = x.y + y ;
       res.z = x.z + y ;
       res.w = x.w + y ;
       return res ;
}

__device__ __forceinline__ double4 operator*(double4 x, float y)
{
       double4 res ;
       res.x = x.x * y ;
       res.y = x.y * y ;
       res.z = x.z * y ;
       res.w = x.w * y ;
       return res ;
}

__device__ __forceinline__ double4 operator*(double4 x, double4 y)
{
       double4 res ;
       res.x = x.x * y.x ;
       res.y = x.y * y.y ;
       res.z = x.z * y.z ;
       res.w = x.w * y.w ;
       return res ;
}

__device__ __forceinline__ double4 __fma4_rn(double4 x, double4 y, float z)
{
       double4 res ;
       res.x = x.x * y.x + z ;
       res.y = x.y * y.y + z ;
       res.z = x.z * y.z + z ;
       res.w = x.w * y.w + z ;
       return res ;
}

// operations count (each)
// + : 1
// * : 2
// *+ : 5
__device__ __forceinline__ double4 myexpm1double4 (double4 x)
{
       double4 res = x + 8.0f ;
       res = __fma4_rn (res, x, 56.0f) ;
       res = __fma4_rn (res, x, 336.0f) ;
       res = __fma4_rn (res, x, 1680.0f) ;
       res = __fma4_rn (res, x, 6720.0f) ;
       res = __fma4_rn (res, x, 20160.0f) ;
       return res * x * 2.4801587301587301587301587301587e-5f ;
}

__device__ __forceinline__ double4  load (float4 x)
{
	double4 res ;
	res.x = x.x ;
	res.y = x.y ;
	res.z = x.z ;
	res.w = x.w ;
	return res ;
}

__device__ __forceinline__ float4  store (double4 x)
{
	float4 res ;
	res.x = x.x ;
	res.y = x.y ;
	res.z = x.z ;
	res.w = x.w ;
	return res ;
}



// 64 * (arraysize * 4) flops minimum for this kernel
__global__ void expkerneldouble4(int arraySize, const double4 * a, double4 * b)
{
       for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < arraySize ; i += blockDim.x * gridDim.x)
       {
             // b [i] = a [i] ;
             b [i] = myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (myexpm1double4 (a [i])))))))))))) ;
       }
}

__device__ __forceinline__ double myexpm1d (double x)
{
       double res = x + 8.0 ;
       res = res * x + 56.0 ;
       res = res * x + 336.0 ;
       res = res * x + 1680.0 ;
       res = res * x + 6720.0 ;
       res = res * x + 20160.0 ;
       return res * x * 2.4801587301587301587301587301587e-5 ;
}

__global__ void expkerneldouble(int arraySize, const double * __restrict a, double* b)
{
       for (int i = threadIdx.x + blockDim.x * blockIdx.x ; i < arraySize ; i += blockDim.x * gridDim.x)
       {
             b [i] =   myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (myexpm1d (a [i])))))))))))) ;
       }
}


void cuda_check_func (cudaError_t er, int line, const char* file)
{
       if (er != cudaSuccess) ::fprintf (stderr, "[%s:%d] -- Cuda error : %d %s\n", file, line, er, ::cudaGetErrorString (er)) ;
}

#define CUDA_CHECK(a) cuda_check_func(a,__LINE__,__FILE__)

int main()
{
      CUDA_CHECK(::cudaSetDevice (0)) ;

       cudaDeviceProp props ;
       CUDA_CHECK (::cudaGetDeviceProperties (&props, 0)) ;

      int nthreads = 1024 ;
      int nblocks = 32 * props.multiProcessorCount ;

       int arraySize = nblocks*128*1024 ;
       float *da, *db, *dc, *dd ;
         double *dda, *ddb, *ddc ;
       CUDA_CHECK (::cudaMalloc ((void**)&da, arraySize * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc ((void**)&db, arraySize * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc ((void**)&dc, arraySize * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc ((void**)&dd, arraySize * sizeof(float))) ;
       CUDA_CHECK (::cudaMalloc ((void**)&dda, arraySize * sizeof(double))) ;
       CUDA_CHECK (::cudaMalloc ((void**)&ddb, arraySize * sizeof(double))) ;
       CUDA_CHECK (::cudaMalloc ((void**)&ddc, arraySize * sizeof(double))) ;
       initkernel <<< nblocks, nthreads >>> (arraySize, da, db, dc, dd, dda, ddb, ddc) ;
       
       // [mharris] need a sync before you start the timer to ensure you are timing only one kernel!
       CUDA_CHECK (::cudaDeviceSynchronize()); 

      StartTimer() ;

       expkernel <<< nblocks, nthreads >>> (arraySize, da, db) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

      double plain = GetTimer() / 1000.0 ;
      StartTimer() ;

       expkernelvect4 <<< nblocks, nthreads >>> (arraySize/4, (float4*)da, (float4*)dc) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

      double v4 = GetTimer() / 1000.0 ;
      StartTimer() ;

       expkernelfloat4 <<< nblocks, nthreads >>> (arraySize/4, (float4*)da, (float4*)dd) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

      double f4 = GetTimer() / 1000.0 ;
      StartTimer() ;

       expkerneldouble <<< nblocks, nthreads >>> (arraySize, dda, ddb) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

      double plaind = GetTimer() / 1000.0 ; 
      StartTimer() ;

       expkerneldouble4 <<< nblocks, nthreads >>> (arraySize/4, (double4*)dda, (double4*)ddc) ;
       CUDA_CHECK (::cudaDeviceSynchronize());

      double d4 = GetTimer() / 1000.0 ;

       float* a = new float[arraySize] ;
       float* b = new float[arraySize] ;
       float* c = new float[arraySize] ;
       float* d = new float[arraySize] ;

       CUDA_CHECK (::cudaMemcpy (a, da, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;
       CUDA_CHECK (::cudaMemcpy (b, db, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;
       CUDA_CHECK (::cudaMemcpy (c, dc, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;
       CUDA_CHECK (::cudaMemcpy (d, dd, sizeof (float) * arraySize, cudaMemcpyDeviceToHost)) ;

       ::fprintf (stdout, "x, op, op4, fl4\n") ;
       for (int k = 0 ; k < 100 ; ++k)
       {
             ::fprintf (stdout, "%f, %f, %f, %f\n", a[k], b[k], c[k], d[k]) ;
       }

       double flopCount = 13.0 * 12.0 * (double)arraySize ;

       ::fprintf (stderr, "regular operation took %lf s\n", plain) ;
       ::fprintf (stderr, "vect4 operation took %lf s\n", v4) ;
       ::fprintf (stderr, "float4 operation took %lf s\n", f4) ;
       ::fprintf (stderr, "double operation took %lf s\n", plaind) ;
       ::fprintf (stderr, "double4 operation took %lf s\n", d4) ;

       ::fprintf (stderr, "regular FLOPs >= %lf GFLOPs\n", (double) flopCount / plain / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "vect4 FLOPs >= %lf GFLOPs\n", (double) flopCount / v4 / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "float4 FLOPs >= %lf GFLOPs\n", (double) flopCount / f4 / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "double FLOPs >= %lf GFLOPs\n", (double) flopCount / plaind / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "double4 FLOPs >= %lf GFLOPs\n", (double) flopCount / d4 / (1000.0*1000.0*1000.0)) ;


      ::fprintf (stderr, "CLOCK RATE = %d\n", props.clockRate) ;
      ::fprintf (stderr, "max threads per sm = %d\n", props.maxThreadsPerMultiProcessor) ;
      
       double theoFlops = 1000.0
                    *   props.clockRate
                    *   props.multiProcessorCount
                    *   192.0  // number of cores per MP 
                    *   2.0 ;  // FMA
       ::fprintf (stderr, "Theoretical GFLOPS = %lf GFLOPs\n", theoFlops / (1000.0*1000.0*1000.0)) ;

       ::fprintf (stderr, "COUNTING FMA AS A SINGLE FLOP AND RESCALING IN INSCRUCTION COUNT\n") ;

       flopCount = 8.0 * 12.0 * (double)arraySize ; //flopCount * 8.0 / 13.0 ;

       ::fprintf (stderr, "regular FLOPs >= %lf GFLOPs\n", (double) flopCount / plain / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "vect4 FLOPs >= %lf GFLOPs\n", (double) flopCount / v4 / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "float4 FLOPs >= %lf GFLOPs\n", (double) flopCount / f4 / (1000.0*1000.0*1000.0)) ;
         ::fprintf (stderr, "double FLOPs >= %lf GFLOPs\n", (double) flopCount / plaind / (1000.0*1000.0*1000.0)) ;
         ::fprintf (stderr, "double4 FLOPs >= %lf GFLOPs\n", (double) flopCount / d4 / (1000.0*1000.0*1000.0)) ;

       theoFlops = theoFlops / 2.0 ;
       ::fprintf (stderr, "Theoretical GFLOPS = %lf GFLOPs\n", theoFlops / (1000.0*1000.0*1000.0)) ;
       ::fprintf (stderr, "VECT4 UTILIZATION = %2.1f percent\n", 100.0f * (double) flopCount / v4 / theoFlops) ;

      return 0;
}


